// $Id: file_path.h 4812 2007-01-30 00:54:20Z jz $

#ifndef __ZSCHIMMER_FILE_PATH_H
#define __ZSCHIMMER_FILE_PATH_H

#include "message.h"


namespace zschimmer {

void                            test_file_path              ();

//-----------------------------------------------------------------------------------------File_path

struct File_path : string
{
    static char                 directory_separator         ( const string& example_path = "" );
    static char                 directory_separator         ( const string& example_path1, const string& example_path2 );


                                File_path                   ()                                      {}
                                File_path                   ( const string& path )                  { set_path( path ); }
                                File_path                   ( const string&    directory, const string& tail_path );

    File_path&                  operator =                  ( const string& path )                  { set_path( path );  return *this; }

    void                    set_name                        ( const string& );                                
    string                      name                        () const;
    void                    set_directory                   ( const string& );
    File_path                   directory                   () const;
    void                        prepend_directory           ( const string& );
    string                      extension                   () const;
    string                      basename                    () const;
    const string&               path                        ()  const                               { return *static_cast<const string*>( this ); }
    void                    set_path                        ( const string& path )                  { *static_cast<string*>( this ) = path; }
  //const string&               path                        () const                                { return _path; }
    string                      base_name                   () const;
    bool                     is_absolute_path               () const;
    string                      absolute_path               () const;
    void                        unlink                      () const;
    bool                        try_unlink                  () const;
    void                        move_to                     ( const File_path& ) const;

  private:
    //string                     _path;
};

inline void                     insert_into_message         ( Message_string* m, int index, const File_path& f ) throw()  { return m->insert_string( index, f.path() ); }

//----------------------------------------------------------------------------------------File_info

struct File_info : Object
{
    static int                  quick_last_write_compare    ( const File_info*, const File_info* ); // last_write() muss vor einmal aufgerufen worden sein!
    static bool                 quick_last_write_less       ( const File_info*, const File_info* ); // last_write() muss vor einmal aufgerufen worden sein!

    
                                File_info                   ()                                      : _zero_( this+1 ) {}

    File_path&                  path                        ()                                      { return _file_path; }
    void                    set_create_time                 ( time_t );
    time_t                      create_time                 ();
    void                    set_last_access_time            ( time_t );
    time_t                      last_access_time            ();
    void                    set_last_write_time             ( time_t );
    time_t                      last_write_time             ();

    bool                        try_call_stat               ();
    void                        call_stat                   ();

  private:
    void                        assert_uint32               ( time_t, const char* function );


    Fill_zero                  _zero_;
    
    // Speicher kompakt fllen (fr 300.000 File_info's). 
    // Position 9
    bool                       _create_time_filled;
    bool                       _last_access_time_filled;
    bool                       _last_write_time_filled;

    // Position 12
    uint32                     _create_time;                // uint32 gilt etwa bis zum Jahr 2106
    uint32                     _last_access_time;           // uint32 gilt etwa bis zum Jahr 2106
    uint32                     _last_write_time;            // uint32 gilt etwa bis zum Jahr 2106

    // Position 24
    File_path                  _file_path;

    // Position 52 (Visual C++ 2005)
};

//------------------------------------------------------------------------------------mini_string<>
#if 0

// Weitere Ersparnis mit File_info_ptr statt ptr<File_info>. Damit entfllt die vtbl. refcnt kann short sein. Oder Byte (mit assert).
// bool in drei Bits zusammenfassen: int _create_time_filled:1;
// Mit mini_string<27> wren das 32 Bytes statt jetzt 52
// Nur fr einen Thread!
// Fr eine handvoll weniger Pointer
// Auf Fill_zero verzichten (-1 Byte)
// Mit dieser Technik kann mini_string 5 Bytes kleiner werden: 1+1+4+4=10 oder (64bit) 1+1+8+4=14

struct simple_single_thread_object< class CLASS >
{
    AddRef() { return assert( _reference_count < BYTE_MAX ); _reference_count++; }
    Release() { if( --_reference_count == 0 )  delete this; }

    Byte    _reference_count;
}


// Speicherverwaltung fr Objekte gleicher Gre braucht weniger Platz fr die Verwaltung, bei groen Seiten
// Kann in Memory mapped files gehalten werden.

struct page< class OBJECT, uint PAGE_SIZE >
{
    OBJECT _object_array [ PAGE_SIZE / sizeof (OBJECT_SIZE) - ( ( ( PAGE_SIZE / sizeof (OBJECT) ) + 7 ) / 8 + sizeof (OBJECT) - 1 ) / sizeof (OBJECT) ) ];
    Byte [ ( ( PAGE_SIZE / sizeof (OBJECT) ) + 7 ) / 8 ] _bit_map;
};

struct Pages
{
    vector<Page*>           _pages;
};


//Etwas Sparsamer als File_path (28 Bytes brutto, 16 (-1?) Byte netto fr internen String

template< uint SIZE >   // Empfehlung: SIZE >= sizeof (void*) + sizeof (int) + 1, 32bit: mini_string<9>, 64bit: mini_string<13>, besser aufgerundet: <16>, <24>
struct mini_string : simple_single_thread_object< mini_string< SIZE> >
{
    mini_string()
    {
        _chars[ SIZE ] = '\0';
        _chars[ 0 ] = '\0';
    }

  private:
    struct Allocated_string;
    {
        char*   _ptr;
        uint32  _size;
    };

    union
    {
        Allocated_string  _allocated_string;
        char              _chars            [ SIZE-1 ];     // Ein Zeichen weniger fr _is_local_string, damit SIZE auf Wortgrenze liegen kann 
    };

    bool    _is_local_string;
};

typedef mini_string< 2 * sizeof (void*) + sizeof (uint32)>  Mini_string;

#endif
//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

#endif
