// $Id: xml.h 4580 2006-10-24 15:45:14Z jz $

#include "zschimmer.h"
#include "z_io.h"
#include "log.h"

namespace zschimmer {
namespace io {

using namespace ::std;

//--------------------------------------------------------------------------------------------const

const size_t Buffered_writer::default_buffer_byte_count = 50000;

//------------------------------------------------------------------------------------Writer::write
    
void Writer::write( char c )
{
    write( string( &c, 1 ) );
}

//--------------------------------ransparent_output_stream_writer::Transparent_output_stream_writer

Transparent_output_stream_writer::Transparent_output_stream_writer( Output_stream* output_stream )
: 
    _output_stream(output_stream ) 
{
}

//----------------------------------------------------------Transparent_output_stream_writer::write
    
void Transparent_output_stream_writer::write( const string& text )
{ 
    _output_stream->write_bytes( text ); 
}

//----------------------------------------------------------Transparent_output_stream_writer::flush

void Transparent_output_stream_writer::flush()
{ 
    _output_stream->flush(); 
}

//-----------------------------------------------------------------Buffered_writer::Buffered_writer

Buffered_writer::Buffered_writer( Writer* writer, int byte_count ) 
: 
    _writer(writer), 
    _buffer_size(byte_count)
{
}

//----------------------------------------------------------------Buffered_writer::~Buffered_writer
    
Buffered_writer::~Buffered_writer()
{
    try
    {
        flush();
    }
    catch( exception& x ) { Z_LOG( __FUNCTION__ << "  " << x << '\n' ); }
}


//---------------------------------------------------------------------------Buffered_writer::write

void Buffered_writer::write( const string& text )
{
    if( _buffer.capacity() < _buffer_size )  _buffer.reserve( _buffer_size );

    size_t length = min( text.length(), _buffer_size - _buffer.length() );
    _buffer.append( text.data(), length );

    if( _buffer.length() == _buffer_size )  flush();

    size_t position = length;
    while( position < text.length() )
    {
        size_t length = min( text.length() - position, _buffer_size - _buffer.length() );
        _buffer.append( text.data() + position, length );
        position += length;
        if( _buffer.length() == _buffer_size )  flush();
    }
}

//---------------------------------------------------------------------------Buffered_writer::flush

void Buffered_writer::flush()
{
    if( _buffer.length() > 0 )  
    {
        _writer->write( _buffer );
        _buffer.erase();
    }
}

//-------------------------------------------------------------------------------------------------

} //namespace io
} //namespace zschimmer
