// $Id: xml.h 4580 2006-10-24 15:45:14Z jz $

#ifndef __ZSCHIMMER_Z_IO_H
#define __ZSCHIMMER_Z_IO_H



namespace zschimmer {
namespace io {

//-------------------------------------------------------------------------------------------Writer

struct Writer : Object
{
    virtual void                write                       ( char );
    virtual void                write                       ( const string& )                       = 0;  
    virtual void                flush                       ()                                      = 0;
};

//------------------------------------------------------------------------------------Output_stream

struct Output_stream : Object
{
    virtual void                write_bytes                 ( const string& bytes )                 = 0;
    virtual void                flush                       ()                                      = 0;
};

//-------------------------------------------------------------------------------------Input_stream

struct Input_stream : Object
{
    virtual string              read_bytes                  ()                                      = 0;
};

//-----------------------------------------------------------------Transparent_output_stream_writer

struct Transparent_output_stream_writer : Writer
{
                                Transparent_output_stream_writer( Output_stream* );
    
    virtual void                write                       ( const string& );
    virtual void                flush                       ();

  private:
    ptr<Output_stream>         _output_stream;
};

//-------------------------------------------------------------------------------------------------

struct Buffered_writer : Writer
{
    static const size_t         default_buffer_byte_count;


                                Buffered_writer             ( Writer*, int default_buffer_byte_count );
                               ~Buffered_writer             ();

    void                        write                       ( const string& );
    void                        flush                       ();

  private:
    ptr<Writer>                _writer;
    string                     _buffer;
    size_t                     _buffer_size;
};

//-------------------------------------------------------------------------------------------------

} //namespace io
} //namespace zschimmer

#endif
